#!/usr/bin/env python3
import os
import argparse
import re
from alara_split_gather.utils import get_num_tasks

def check_status(sep='_', prefix='phtn_src', threshold_size=5.0): # generate filenams for checking
    files = []
    sizes = []
    running = []
    finished = []
    # get the folders start with task*
    num_tasks = get_num_tasks()
    print(f"{num_tasks} sub-task folders found")
    for i in range(num_tasks):
        filename = os.path.join(".", f"task{i}", f"{prefix}{sep}{i}")
        files.append(filename)
        size = os.path.getsize(filename)
        sizes.append(size)
        # count running/finshed status 
        if size > threshold_size:
            finished.append(i)
        else:
            running.append(i)

    # print the summary
    # finished tasks
    fin_str = f"{len(finished)} sub-tasks are finished (with output greater than {threshold_size} bytes, they are:\n     "
    count = 0
    for i in finished:
        count += 1
        fin_str = f"{fin_str} {i}"
        if count>0  and count%10 == 0:
            fin_str = f"{fin_str}\n     "
    print(fin_str)
    # running tasks
    run_str = f"{len(running)} sub-tasks are running, they are:\n     "
    count = 0
    for i in running:
        count += 1
        run_str = f"{run_str} {i}"
        if count>0 and count%10 == 0:
            run_str = f"{run_str}\n     "
    print(run_str)

   
def alara_tasks_status():
    check_tasks_status_help = ('This script check the status of alara tasks\n')
    parser = argparse.ArgumentParser()
    parser.add_argument("-s", "--separator", required=False, help=" '_' or '-'")
    parser.add_argument("-p", "--prefix", required=False, help="prefix of phtn_src")
    parser.add_argument("-t", "--threshold_size", required=False, help="threshold file size [byte]")
    args = vars(parser.parse_args())

    # prefix
    prefix = "phtn_src"
    if args['prefix'] is not None:
        prefix = args['prefix']

    # separator
    sep = '_'
    if args['separator'] is not None:
        if args['separator'] not in ['_', '-']:
            raise ValueError(f"separator {args['separator']} not supported!")
        sep = args['separator']

    # threhold size
    threshold_size = 5.0
    if args['threshold_size'] is not None:
        try:
            threshold_size = float(args['threshold_size']) 
        except:
            raise ValueError(f"threshold size must be a float/int number!")

    check_status(sep=sep, prefix=prefix, threshold_size=threshold_size)

if __name__ == '__main__':
    alara_tasks_status()



#!/usr/bin/env python3
import numpy as np
import tables as tb
import os
import argparse
from pyne.alara import _make_response_dtype
from alara_split_gather.utils import get_num_tasks

def append_photon_source_to_hdf5(num_tasks=2, nucs='all', chunkshape=(10000,),
        output='phtn_src.h5',sep='_'):
    """Converts a plaintext photon source file to an HDF5 version for
    quick later use.

    This function produces a single HDF5 file named <filename>.h5 containing the
    table headings:

        idx : int
            The volume element index assuming the volume elements appear in xyz
            order (z changing fastest) within the photon source file in the case of
            a structured mesh or mesh.mesh_iterate() order for an unstructured mesh.
        nuc : str
            The nuclide name as it appears in the photon source file.
        time : str
            The decay time as it appears in the photon source file.
        phtn_src : 1D array of floats
            Contains the photon source density for each energy group.

    Parameters
    ----------
    nucs : str
        Nuclides need to write into h5 file. For example:
            - 'all': default value. Write the information of all nuclides to h5.
            - 'total': used for r2s. Only write TOTAL value to h5.
    num_tasks : int
        Number of alara tasks.
    chunkshape : tuple of int
        A 1D tuple of the HDF5 chunkshape.
    """
    # create a hdf5 file
    phtn_dtype = _make_response_dtype('phtn_src', data_length=24)
    filters = tb.Filters(complevel=1, complib='zlib')
    h5f = tb.open_file(output, 'w', filters=filters)
    tab = h5f.create_table('/', 'data', phtn_dtype, chunkshape=chunkshape)

    # read phtn_src
    chunksize = chunkshape[0]
    rows = np.empty(chunksize, dtype=phtn_dtype)
    idx = 0
    old = ""
    row_count = 0
    files = []
    # get phtn_src files
    print(f"{num_tasks} files to be write into {output}")
    for i in range(num_tasks):
        filename = os.path.join(".", f"task{i}", f"phtn_src{sep}{i}")
        files.append(filename)
    for i, filename in enumerate(files):
        print(f"    running with file {filename}, [{i}/{num_tasks}]")
        f = open(filename, 'r')
        for i, line in enumerate(f, 1):
            tokens = line.strip().split('\t')
            # Keep track of the idx by delimiting by the last TOTAL line in a
            # volume element.
            if tokens[0] != u'TOTAL' and old == u'TOTAL':
                idx +=1
            if nucs.lower() == 'all':
                row_count += 1
                j = (row_count-1)%chunksize
                rows[j] = (idx, tokens[0].strip(), tokens[1].strip(),
                    np.array(tokens[2:], dtype=np.float64))
            elif nucs.lower() == 'total':
                if tokens[0] == 'TOTAL':
                    row_count += 1
                    j = (row_count-1)%chunksize
                    rows[j] = (idx, tokens[0].strip(), tokens[1].strip(),
                        np.array(tokens[2:], dtype=np.float64))
            else:
                h5f.close()
                f.close()
                raise ValueError(u"Nucs option {0} not support!".format(nucs))

            # Save the nuclide in order to keep track of idx
            old = tokens[0]
            if (row_count>0) and (row_count%chunksize == 0):
                tab.append(rows)
                rows = np.empty(chunksize, dtype=phtn_dtype)
                row_count = 0
        f.close()

    if row_count % chunksize != 0:
        tab.append(rows[:j+1])
    h5f.close()
    print("Done")

def merge_phtn_src(num_tasks=2, prefix="phtn_src", output="phtn_src", sep='_'):
    """
    Merge phtn_src of sub-tasks into a single phtn_src.
    """
    print(f"{num_tasks} files to merge.")
    # get phtn_src files
    files = []
    for i in range(num_tasks):
        filename = os.path.join(".", f"task{i}", f"{prefix}{sep}{i}")
        files.append(filename)
    # write content
    fo = open(output, 'w')
    for i, filename in enumerate(files):
        print(f"    merging {filename}, [{i}/{num_tasks}]")
        with open(filename, 'r') as fin:
            cnt = fin.read()
            fo.write(cnt)
    fo.close()

def alara_gather_tasks():
    gather_alara_task_help = ('This script gather alara output and phtn_src\n')
    parser = argparse.ArgumentParser()
    parser.add_argument("-m", "--mode",required=False, help="mode of gather, txt or h5")
    parser.add_argument("-p", "--prefix", required=False, help="prefix of alara output")
    parser.add_argument("-s", "--separator", required=False, help=" '_' or '-'")
    args = vars(parser.parse_args())

    # mode
    mode = 'h5'
    if args['mode'] is not None:
        mode = args['mode']
    print(f"mode: {mode}")
    
    # prefix
    prefix = "phtn_src"
    if args['prefix'] is not None:
        prefix = args['prefix']
    print(f"prefix: {prefix}")

    # separator
    sep = '_'
    if args['separator'] is not None:
        if args['separator'] not in ['_', '-']:
            raise ValueError(f"separator {args['separator']} not supported!")
        sep = args['separator']
    print(f"separator: {sep}")

    # output filename
    if mode == 'h5':
        output = f"{prefix}.h5"
    elif mode == 'txt':
        output = f"{prefix}"
    print(f"output file name: {output}")
 
    # get num_tasks
    num_tasks = get_num_tasks()
    print(f"{num_tasks} sub-tasks are found in {os.path.abspath('.')}")

    # append/merge the file
    if mode == 'h5':
        append_photon_source_to_hdf5(num_tasks=num_tasks, nucs='TOTAL',
            output=output, sep=sep)
    elif mode == 'txt':
        merge_phtn_src(num_tasks=num_tasks, prefix=prefix, output=output, sep=sep)
    else:
        raise ValueError("Wrong mode")


if __name__ == '__main__':
    alara_gather_tasks()



#!/usr/bin/env python3
import re
import os
import argparse
from alara_split_gather.utils import assign_subtasks

def is_volume_start(line):
    volume_start_pattern = re.compile("^volume", re.IGNORECASE) 
    if re.match(volume_start_pattern, line):
        return True
    else:
        return False 
def is_end(line):
    end_pattern = re.compile("^end", re.IGNORECASE)
    if re.match(end_pattern, line):
        return True
    else:
        return False

def is_mat_loading(line):
    mat_loading_pattern = re.compile("^mat_loading")
    if re.match(mat_loading_pattern, line):
        return True
    else:
        return False

def count_vols(inp="alara_inp"):
    """Count the number of vols"""
    count = 0
    with open(inp, 'r') as fin:
        vol_start = False
        while True:
            line = fin.readline()
            if not line:
                break
            if is_volume_start(line):
                vol_start = True
            if vol_start and (not is_end(line)) and (not is_volume_start(line)):
                count += 1
            if vol_start and is_end(line):
                return count

def get_zones_vols(inp="alara_inp"):
    zones, vols = [], []
    with open(inp, 'r') as fin:
        vol_start = False
        while True:
            line = fin.readline()
            if not line:
                break
            if is_volume_start(line):
                vol_start = True
            if vol_start and (not is_end(line)) and (not is_volume_start(line)):
                line_ele = line.strip().split()
                vols.append(line_ele[0])
                zones.append(line_ele[1])
            if is_end(line) and vol_start:
                break
    return zones, vols

def get_mats(inp="alara_inp"):
    mats = []
    with open(inp, 'r') as fin:
        mat_loading_start = False
        while True:
            line = fin.readline()
            if not line:
                break
            if is_mat_loading(line):
                mat_loading_start = True
            if mat_loading_start and (not is_mat_loading(line)) and (not is_end(line)):
                line_ele = line.strip().split()
                mats.append(line_ele[1])
            if mat_loading_start and is_end(line):
                break
    return mats

def split_alara_inp(inp="alara_inp", num_tasks=10, sep='_', truncation=None):
    """Split alara_inp into num_tasks sub-tasks"""
    num_vols = count_vols(inp)
    if num_vols < num_tasks:
        raise ValueError(f"num_tasks must be smaller than {num_vols}")
    zones, vols = get_zones_vols(inp)
    mats = get_mats(inp)
    #print(zones, vols, mats)
    vol_per_task = num_vols // num_tasks
    subtasks = assign_subtasks(num_vols, num_tasks)
    for tid in range(len(subtasks)):
        print(f"write files for task {tid}")
        inp_name = f"{inp}{sep}{tid}"
        fin = open(inp, 'r')
        fo = open(inp_name, 'w')
        #lines = fin.readlines()
        vol_start, vol_end = False, False
        mat_start, mat_end = False, False
        vol_write, mat_write = False, False
        while True:
            line = fin.readline()
            if ("" == line):
                break
            if is_volume_start(line):
                vol_start = True
                fo.write(line)
                continue
            if (not vol_start):
                fo.write(line)
            if vol_start and (not vol_end) and (not vol_write):
                # write specific vols
                #for i in range(vol_per_task):
                for i, item in enumerate(subtasks[tid]):
                    vid = tid*vol_per_task + i
                    cnt = f"\t {vols[vid]} {zones[vid]}\n"
                    fo.write(cnt)
                vol_write = True
                continue
            if is_end(line) and vol_start:
                vol_end = True
            if is_mat_loading(line):
                mat_start = True
                fo.write(line)
                continue
            if vol_end and not mat_start:
                fo.write(line)
            if mat_start and (not mat_end) and (not mat_write):
                # write specific mats
                #for i in range(vol_per_task):
                for i, item in enumerate(subtasks[tid]):
                    vid = tid*vol_per_task + i
                    cnt = f"\t {zones[vid]} {mats[vid]}\n"
                    fo.write(cnt)
                mat_write = True
                continue
            if mat_start and is_end(line):
                mat_end = True
            # mixture part
            if "mixture" in line:
                #import pdb; pdb.set_trace()
                line_ele = line.strip().split()
                if line_ele[1] in mats[tid*vol_per_task:tid*vol_per_task+len(subtasks[tid])]:
                    fo.write(line)
                    while True:
                        line = fin.readline()
                        if "end" in line:
                            fo.write(line)
                            break
                        else:
                            fo.write(line)
                else:
                    while True: # skip these lines
                        line = fin.readline()
                        if "end" in line:
                            line = fin.readline()
                            break
                continue
            if "phtn_src" in line:
                line_ele = line.strip().split()
                phtn_str = "     "
                for i in range(len(line_ele)):
                    if "phtn_src" in line_ele[i]:
                        line_ele[i] = f"{line_ele[i]}{sep}{tid}"
                    phtn_str = f"{phtn_str} {line_ele[i]}"
                fo.write(phtn_str)
                fo.write("\n")
                continue
            if "alara_fluxin" in line:
                #ctn = line.replace("alara_fluxin", f"alara_fluxin{tid}")
                line_ele = line.strip().split()
                skip = int(line_ele[4])
                skip += tid*vol_per_task
                #ctn = line.replace(" 0 ", f" {skip}")
                ctn = f"{line_ele[0]} {line_ele[1]} {line_ele[2]} {line_ele[3]} {skip} default"
                fo.write(ctn)
                fo.write("\n")
                continue
            if "truncation" in line and truncation is not None:
                line_ele = line.strip().split()
                ctn = f"{line_ele[0]} {truncation}\n"
                fo.write(ctn)
                continue
            if mat_end:
                fo.write(line)
        fo.close()
        fin.close()

def pbs_generation(tid, inp="alara_inp", sep='_'):
    """Generate a pbs file for current task id"""
    fo = open(f"alara_task{sep}{tid}.pbs", 'w')
    fo.write("#!/bin/bash\n")
    fo.write(f"#PBS -N task{sep}{tid}\n")
    fo.write("#PBS -l nodes=1:ppn=28\n")
    fo.write("cd $PBS_O_WORKDIR\n")
    fo.write(f"$HOME/opt/ALARA/bin/alara {inp}{sep}{tid} > output{sep}{tid}.txt\n")
    fo.close()

def alara_split_task():
    """
    Split the alara task
    """
    split_alara_task_help = ('This script read an alara_inp and split it into samller task\n')
    parser = argparse.ArgumentParser()
    parser.add_argument("-n", "--num_tasks", required=False, help="number to sub-tasks, default: 2")
    parser.add_argument("-i", "--input", required=False, help="ALARA input, default: alara_inp")
    parser.add_argument("-d", "--data", required=False, help="ALARA data file, default: ./data")
    parser.add_argument("-s", "--separator", required=False, help=" '_' or '-'")
    parser.add_argument("-t", "--truncation", required=False, help="reset turncation")
    args = vars(parser.parse_args())

    # input
    inp = "alara_inp"
    if args['input'] is not None:
        inp = args['input']

    # number of tasks
    if args['num_tasks'] is not None:
        num_tasks = int(args['num_tasks'])
        print(f"{num_tasks} sub-tasks will be generated")

    # separator
    sep = '_'
    if args['separator'] is not None:
        if args['separator'] not in ['_', '-']:
            raise ValueError(f"separator {args['separator']} not supported!")
        sep = args['separator']

    # alara data
    data = "./data"
    if args['data'] is not None:
        data = args['data']
        print(f"ALARA data: {data}")

    # truncation
    truncation = 1e-12
    if args['truncation'] is not None:
        truncation = float(args['truncation'])
        print(f"use truncation: {truncation}")

    split_alara_inp(inp, num_tasks, sep=sep, truncation=truncation)
    
    # setup sub-task directories
    for i in range(num_tasks):
        os.system(f"mkdir -p task{i}")
        os.system(f"ln -sf {os.path.abspath(data)} task{i}/data")
        # goes into sub-task dir
        print(f"copy/link files for task {i}")
        os.chdir(f"task{i}")
        os.system(f"mv ../{inp}{sep}{i} .")
        os.system(f"ln -sf ../alara_fluxin alara_fluxin")
        os.system(f"ln -sf ../alara_matlib alara_matlib")
        #pbs_generation(i, inp=inp, sep=sep)
        os.chdir("..")
        

if __name__ == '__main__':
    alara_split_task()



from setuptools import setup, find_packages
 
setup(name='alara_split_gather',
      version='0.1.8',
      url='https://github.com/zxkjack123/alara_split_gather',
      license='MIT',
      author='Xiaokang Zhang',
      author_email='zxkjack123@163.com',
      description='Python package to split and gather ALARA tasks',
      packages=find_packages(exclude=['tests']),
      package_dir={'alara_split_gather':'alara_split_gather'},
      include_package_data=True,
      package_data={'alara_split_gather':[
                        ],
                    },
      long_description=open('README.md').read(),
      entry_points={
          "console_scripts":["alara_split_task = alara_split_gather.alara_split_task:alara_split_task",
                             "alara_tasks_status = alara_split_gather.check_tasks_status:alara_tasks_status",
                             "alara_gather_tasks = alara_split_gather.alara_gather_tasks:alara_gather_tasks"
                             ]
          },
      classifiers=[
          "Programming Language :: Python :: 3",
          "License :: OSI Approved :: MIT License",
          "Operating System :: OS Independent",
          "Environment :: Console",
          ],
      python_requires='>=3.6',
      zip_safe=False)



#!/usr/bin/env python3
import os
import re

def assign_subtasks(total_num, num_tasks=2):
    """
    Split the task into sub-tasks.

    Parameters:
    -----------
    total_num: int
        Total number of the task to be split.
    num_tasks: int
        The number of sub-tasks.

    Returns:
    --------
    subtask_ids: list of list
        List of the list of ids. Eg.
        [[0, 1], [2, 3, 4]]
    """
    subtasks = []
    num_per_task = total_num // num_tasks
    for i in range(num_tasks):
        if i < num_tasks-1:
            ids = list(range(i*num_per_task, (i+1)*num_per_task, 1))
        else:
            ids = list(range(i*num_per_task, total_num))
        subtasks.append(ids)
    return subtasks
        
def diff_check_file(f1, f2):
    command = ''.join(["diff ", "--strip-trailing-cr ", f1, " ", f2])
    flag = os.system(command)
    return flag

def get_num_tasks():
    # get the folders start with task*
    files = os.listdir(os.path.abspath("."))
    task_pattern = re.compile("^task*")
    num_tasks = 0
    for f in files:
        if re.match(task_pattern, f):
            num_tasks += 1
    return num_tasks
